__author__ = "Markus Pichler"
__credits__ = ["Markus Pichler"]
__maintainer__ = "Markus Pichler"
__email__ = "markus.pichler@tugraz.at"
__version__ = "0.1"
__license__ = "MIT"

from . import IntensityDurationFrequencyAnalyse

IntensityDurationFrequencyAnalyse.command_line_tool()
